//**********************************************************************
// ShowRegACL
// ==========
//
// Show the ACLs for a chosen registry key and all subkeys
//
// You will need the CSecurableObject to
// compile this. They can be downloaded from www.winsrc.freeserve.co.uk.
//
// J. Rennie
// 1st July 2007
//**********************************************************************

#ifndef STRICT
#define STRICT
#endif

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <CSecurableObject/CRegSecObject.h>


//**********************************************************************
// Function prototypes
// -------------------
//**********************************************************************

BOOL ShowRegACL(const char* RegKey);
HKEY OpenKey(const char* RegKey);
BOOL GetSubKey(HKEY hKey, int KeyNum, const char* ParentName, char* SubKeyName);

const char* GetLastErrorMessage(void);
const char* GetLastErrorMessage(DWORD d);


//**********************************************************************
// Structure to help with handling ACEs
// ------------------------------------
//**********************************************************************

#define MAX_USERNAME 32

#define RHSACE_NCINHERIT      0x01 // Child objects that are not containers inherit permissions
#define RHSACE_INHERIT_PASS   0x02 // Child objects inherit and pass on permissions
#define RHSACE_INHERIT_NOPASS 0x04 // Child objects inherit but do not pass on permissions
#define RHSACE_NOINHERIT_PASS 0x08 // Object is not affected by but passes on permissions
#define RHSACE_INHERITED      0x10 // Permissions have been inherited

typedef struct _RhsACE
{
  char  domain[MAX_USERNAME+1],   // This identifies the user
        username[MAX_USERNAME+1];

  DWORD mask, // access mask
        type, // 0 = access allowed, 1 = access denied
        flags;

} RhsACE;


//**********************************************************************
// Global variables
// ----------------
//**********************************************************************

#define SYNTAX "ShowRegACL [-i -x<username>] <keyname>\n" \
               "NB the key must start with HKEY_LOCAL_MACHINE etc\n" \
               "e.g. HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\n" \
               "(the case of the key name doesn't matter)\n"

BOOL IncludeInherited;

#define MAX_EXCLUDE 16
char Exclude[MAX_EXCLUDE][MAX_USERNAME+1];
int  NumExclude;


//**********************************************************************
// main
// ----
//**********************************************************************

int main(int argc, char* argv[])
{ int argnum, i;
  char keyname[1024];

// Process command line flags

  IncludeInherited = FALSE;
  NumExclude = 0;
  argnum = 1;

  while (argnum < argc)
  { if (argv[argnum][0] != '-')
      break;

    AnsiLower(argv[argnum]+1);

    switch (argv[argnum][1])
    { case '?':
        printf(SYNTAX);
        return 0;

      case 'i': // Include inherited permissions
        IncludeInherited = TRUE;
        break;

      case 'x': // Exclude a username
        if (NumExclude <= MAX_EXCLUDE)
        { strcpy(Exclude[NumExclude], argv[argnum]+2);
          NumExclude++;
        }
        break;

      default:
        printf("Unknown flag \"%s\".\n%s", argv[argnum], SYNTAX);
        return(1);
    }

    argnum++;
  }

// Check there is one argument left

  if (argc - argnum < 1)
  { printf(SYNTAX);
    return(2);
  }

// Split the hive off the key

  for (i = 0; argv[argnum][i] != '\\' && argv[argnum][i] != '\0'; i++)
    keyname[i] = argv[argnum][i];
  keyname[i] = '\0';

  if (argv[argnum][i] == '\\')
    i++;

  if (strcmpi(keyname, "HKEY_LOCAL_MACHINE") == 0)
  { strcpy(keyname, "MACHINE");
  }
  else if (strcmpi(keyname, "HKEY_CLASSES_ROOT") == 0)
  { strcpy(keyname, "CLASSES_ROOT");
  }
  else if (strcmpi(keyname, "HKEY_USERS") == 0)
  { strcpy(keyname, "USERS");
  }
  else if (strcmpi(keyname, "HKEY_CURRENT_USER") == 0)
  { strcpy(keyname, "CURRENT_USER");
  }
  else
  { fprintf(stderr, "Unrecognised hive \"%s\"\n%s\n", keyname, SYNTAX);
    return 2;
  }

  strcat(keyname, "\\");
  strcat(keyname, argv[argnum] + i);

// Show the security for the directory and subdirectories

  ShowRegACL(keyname);

// All done

  return 0;
}


//**********************************************************************
// ShowRegACL
// ----------
//**********************************************************************

BOOL ShowRegACL(const char* RegKey)
{ int num_entries, num_aces, i, j;
  BOOL found_ace;
  DWORD mask, type, flags;
  HKEY hkey;
  CRegSecObject sec;
  RhsACE ace[32];
  char username[256], domain[256], rights[16];
  char subkey[1024];
  char outstr[0x1000];

  sprintf(outstr, "%s\n", RegKey);
  found_ace = FALSE;

// Attach the CSecurableObject to this directory

  sec.SetKeyName((char*) RegKey);
  if (!sec.GetObjectSecurity())
  { fprintf(stderr, "Cannot open the key %s: %s\n", RegKey, GetLastErrorMessage(sec.m_iSecErrorCode));
    return FALSE;
  }

// Loop over all ACEs in the list

  num_entries = sec.GetNumEntries();
  num_aces = 0;

  for (i = 0; i < num_entries; i++)
  { sec.GetRightsFromACE(username, domain, &mask, &type, &flags, i);

// Look to see if this is a new user

    for (j = 0; j < num_aces; j++)
      if (strcmp(username, ace[j].username) == 0 && type == ace[j].type)
        break;

// If this is a new user get all their entries

    if (j >= num_aces)
    { strcpy(ace[num_aces].domain, domain);
      strcpy(ace[num_aces].username, username);
      ace[num_aces].type = type;
      ace[num_aces].flags = flags;

// This gets all entries in the ACL matchine the current SID and type
      sec.GetAllRightsFor(&(ace[num_aces].mask), i);

      num_aces++;
    }
  }

// Got all the entries so print them

  for (i = 0; i < num_aces; i++)
  {

// We only want non-inherited permissions unless we've specifically 
// asked for them.

    if (ace[i].flags & RHSACE_INHERITED)
      if (!IncludeInherited)
        continue;

// For tidiness omit the standard "domains" BUILTIN and NT AUTHORITY

    if (strlen(ace[i].domain) == 0 || strcmp(ace[i].domain, "BUILTIN") == 0 || strcmp(ace[i].domain, "NT AUTHORITY") == 0)
      strcpy(username, ace[i].username);
    else
      sprintf(username, "%s\\%s", ace[i].domain, ace[i].username);
    AnsiLower(username);

// Check if we're excluding this username

    for (j = 0; j < NumExclude; j++)
    { if (strcmp(Exclude[j], username) == 0)
        break;
    }

    if (j < NumExclude)
      continue;

// Print the access including whether it's an allowed or denied ACE

    found_ace = TRUE;

    sec.RightsToText(ace[i].mask, rights);
    if (ace[i].flags & RHSACE_INHERITED)
      strcat(rights, "(I)");

    if (ace[i].type == 0)
      sprintf(outstr + strlen(outstr), "\t%s\t%s\n", username, rights);
    else
      sprintf(outstr + strlen(outstr), "\t%s\tDenied: %s\n", username, rights);
  }

// Only print the output if we found a non-inherited ACL

  if (found_ace)
    printf(outstr);

// Now work through subdirectories

  hkey = OpenKey(RegKey);

  if (hkey == INVALID_HANDLE_VALUE)
  { fprintf(stderr, "Cannot open subkeys of %s\n", RegKey, GetLastErrorMessage());
    return 2;
  }

  i = 0;

  while (GetSubKey(hkey, i, RegKey, subkey))
  { i++;

    ShowRegACL(subkey);
  }

  CloseHandle(hkey);

// All done

  return TRUE;
}


//**********************************************************************
// OpenKey
// -------
// Open the named key.
// The name is in the format used by the GetNamedSecurityInfo function
// e.g. "MACHINE\keyname"
//**********************************************************************

HKEY OpenKey(const char* RegKey)
{ int i;
  LONG d;
  HKEY hivekey, h;
  char hive[256], keyname[1024];

// Split the hive off the key

  for (i = 0; RegKey[i] != '\\' && RegKey[i] != '\0'; i++)
    hive[i] = RegKey[i];
  hive[i] = '\0';

  if (RegKey[i] == '\\')
    i++;

  strcpy(keyname, RegKey + i);

  if (strcmpi(hive, "MACHINE") == 0)
  { hivekey = HKEY_LOCAL_MACHINE;
  }
  else if (strcmpi(hive, "CLASSES_ROOT") == 0)
  { hivekey = HKEY_CLASSES_ROOT;
  }
  else if (strcmpi(hive, "USERS") == 0)
  { hivekey = HKEY_USERS;
  }
  else if (strcmpi(hive, "CURRENT_USER") == 0)
  { hivekey = HKEY_CURRENT_USER;
  }
  else
  { fprintf(stderr, "Unrecognised hive \"%s\"\n", hive);
    return (HKEY) INVALID_HANDLE_VALUE;
  }

// Open the key

  d = RegOpenKeyEx(hivekey, keyname, 0, KEY_READ, &h);

  if (d != ERROR_SUCCESS)
    h = (HKEY) INVALID_HANDLE_VALUE;

// Return the key

  return h;
}


//**********************************************************************
// GeySubKey
// ---------
// Get the subkey by number. Return the subkey name in the form used by
// the GetNamedSecurityInfo function e.g. "MACHINE\keyname"
//**********************************************************************

BOOL GetSubKey(HKEY hKey, int KeyNum, const char* ParentName, char* SubKeyName)
{ DWORD subkeysize, d;
  FILETIME f;
  char subkey[256];

// Get the subkey

  subkeysize = 255;
  d = RegEnumKeyEx(hKey, KeyNum, subkey, &subkeysize, NULL, NULL, NULL, &f);

// If there were no more subkeys just return FALSE

  if (d == ERROR_NO_MORE_ITEMS)
    return FALSE;

// If some other error occurred print the error message

  if (d != ERROR_SUCCESS)
  { fprintf(stderr, "Error getting subkey %i of %s\n%s\n", KeyNum, ParentName, GetLastErrorMessage());
    return FALSE;
  }

// Otherwise set the full subkey name

  strcpy(SubKeyName, ParentName);
  strcat(SubKeyName, "\\");
  strcat(SubKeyName, subkey);

// And return indicating success

  return TRUE;
}


//**********************************************************************
// GetLastErrorMessage
// -------------------
//**********************************************************************

const char* GetLastErrorMessage(void)
{ DWORD d;
  static char errmsg[1024];

  d = GetLastError();

  strcpy(errmsg, "<unknown error>");
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, d, 0, errmsg, 1023, NULL);
  return(errmsg);
}


const char* GetLastErrorMessage(DWORD d)
{ static char errmsg[1024];

  strcpy(errmsg, "<unknown error>");
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, d, 0, errmsg, 1023, NULL);
  return(errmsg);
}


